/* An implementation of the non-preemptive, non-work-conserving
 * periodic EDF-CW policy. */

#include "Arduino.h"

#include "tasks.h"
#include "sched_time.h"

static unsigned long next_release_time[MAX_TASKS];
static unsigned long absolute_deadline[MAX_TASKS];
static unsigned long next_absolute_deadline[MAX_TASKS];
static bool pending[MAX_TASKS];

static int compare_next_job_deadlines(const void* task_a, const void* task_b)
{
	const byte *a, *b;

	a = (const byte*) task_a;
	b = (const byte*) task_b;

	/* if we return a negative result, a is sorted before b */
	if (time_after(next_absolute_deadline[*b], next_absolute_deadline[*a]))
		return -1;
	else
		return  1;
}

void edf_cw_init(void)
{
	for (int i = 0; i < MAX_TASKS; i++) {
		next_release_time[i] = 0UL;
		pending[i] = false;
		absolute_deadline[i] = 0UL;
		next_absolute_deadline[i] = tasks[i].deadline;
	}
}

static void edf_cw_release_job(byte i)
{
	absolute_deadline[i] = next_absolute_deadline[i];
	next_absolute_deadline[i] += tasks[i].period;
	next_release_time[i] += tasks[i].period;
	pending[i] = true;
}

sched_decision_t edf_cw_schedule(void)
{
	unsigned long now = current_sched_time();
	char earliest = -1;

	byte tasks_without_pending_jobs[MAX_TASKS];
	byte *nonpending_tail = tasks_without_pending_jobs;

	for (byte i = 0; i < number_of_tasks; i++)
	{
		if (!pending[i] && time_after(now, next_release_time[i])) {
			/* a job of this task has been released */
			edf_cw_release_job(i);
		}
		if (pending[i] &&
		    (earliest < 0 || time_after(absolute_deadline[earliest],
		                                absolute_deadline[i]))) {
			/* consider this job for scheduling */
 			earliest = i;
		}
		if (!pending[i]) {
			/* add to list of tasks without a pending job */
			*nonpending_tail++ = i;
		}
	}


	bool must_idle;
	byte n = nonpending_tail - tasks_without_pending_jobs;

	if (earliest < 0) {
		/* No job is ready. Just idle. */
		must_idle = true;
	} else if (n == 0) {
		/* no tasks without a pending job => just go */
		must_idle = false;
	} else {
		/* some tasks are ready, some are not pending => check if we
		 * need to insert idle time */

		/* Step 1: sort the "next" job of each task by their deadlines */
		qsort(tasks_without_pending_jobs, n, sizeof(byte),
			compare_next_job_deadlines);

		/* Step 2: from the latest deadline, calculate the latest start
		 * time (LST) of that job as follows:
		 *       lst_i = min(lst_{i+1}, dl_i) - C_i where
		 * lst_i is the LST of the i^th job considering that the jobs
		 * are sorted by their deadline. */
		byte last_idx = tasks_without_pending_jobs[n - 1];
		unsigned long lst =
			next_absolute_deadline[last_idx] - tasks[last_idx].wcet;
		for (char i = n - 2; i >= 0; i--) {
			byte idx = tasks_without_pending_jobs[i];
			if (time_after(lst, next_absolute_deadline[idx]))
				lst = next_absolute_deadline[idx] - tasks[idx].wcet;
		 	else
				lst = lst - tasks[idx].wcet;
		}

		/* Step 3: if the WCET of the current highest priority job (EDF
		 * policy) is smaller than or equal to $lst_1 - NOW$ then
		 * schedule the current highest priority job, otherwise wait
		 * until "the earliest next release time". */
		must_idle = time_after(now + tasks[earliest].wcet, lst);
	}

	if (must_idle)
		return NO_JOB_READY;
	else {
		/* schedule the job with the earliest deadline */
		pending[earliest] = false;
		return (sched_decision_t)
			{tasks[earliest].job_function, tasks[earliest].wcet};
	}
}

