/* An implementation of non-preemptive, non-work-conserving,
 * periodic offline-equivalence scheduling.
 */

#include "Arduino.h"

#include "config.h"
#include "panic.h"
#include "tasks.h"
#include "sched_time.h"

#include "sched_oe.h"

#ifdef TABLES_IN_FLASH

/* include statically generated tables */
#include "gentab_oe.h"

/* accessor macros --- get from PGM */
#define get_idle_time(__idx) \
	((struct idle_time_record) \
		{pgm_read_dword_near(&idle_times[__idx].start),	\
		pgm_read_word_near(&idle_times[__idx].duration)})
#define get_irregularity(__tsk, __idx) \
	((struct irregularity_record) \
		{pgm_read_word_near(&irregs[__tsk][__idx].job_id),	\
		pgm_read_dword_near(&irregs[__tsk][__idx].delay)})

#else

static idle_time_record *idle_times = NULL;
static unsigned int idle_times_len;

static irregularity_record* irregs[MAX_TASKS] = {NULL};
static unsigned int irregs_len[MAX_TASKS];

/* accessor macros --- direct memory access */
#define get_idle_time(__idx) idle_times[__idx]
#define get_irregularity(__tsk, __idx) irregs[__tsk][__idx]

static unsigned long hyperperiod;

void oe_set_idle_times(struct idle_time_record *recs, unsigned int count)
{
	idle_times = recs;
	idle_times_len = count;
}

void oe_set_irregularities_for(byte task_id,
	struct irregularity_record *recs, unsigned int count)
{
	irregs[task_id] = recs;
	irregs_len[task_id] = count;
}

void oe_set_hyperperiod(unsigned long hp)
{
	hyperperiod = hp;
	BUG_ON(!hyperperiod);
}

#endif

static unsigned int idle_times_idx;
static unsigned int irregs_idx[MAX_TASKS];

static unsigned long next_arrival_time[MAX_TASKS];
static unsigned long next_inter_arrival_time[MAX_TASKS];
static unsigned int next_job_id[MAX_TASKS];
static byte prio_boosted_tasks[MAX_TASKS + 1];

void oe_init(void)
{
	for (int i = 0; i < MAX_TASKS; i++) {
		// FIXME: what if a task has an irregular arrival in the first
		//        job? Should be handled here. For now, every task
		//        releases its first job at time zero.
		next_arrival_time[i] = 0UL;
		next_inter_arrival_time[i] = tasks[i].period;
		next_job_id[i] = 0UL;
#ifndef TABLES_IN_FLASH
		irregs[i] = NULL;
		irregs_len[i] = 0;
#endif
		irregs_idx[i] = 0;

		prio_boosted_tasks[i] = 0;
	}

	prio_boosted_tasks[MAX_TASKS] = 0;
	idle_times_idx = 0;
#ifndef TABLES_IN_FLASH
	idle_times = NULL;
	idle_times_len = 0;

	hyperperiod = 1000000UL; /* default to something non-zero */
#endif
}

static void oe_hyperperiod_crossing(void)
{
	/* sanity check: make sure we've seen all idle times */
	BUG_ON(idle_times_idx != idle_times_len);
	idle_times_idx = 0;

	for (int i = 0; i < number_of_tasks; i++) {
		/* sanity check: make sure we've een all irregularities */
		BUG_ON(irregs_idx[i] != irregs_len[i]);
		irregs_idx[i] = 0;
		next_job_id[i] = 0;
		next_arrival_time[i] = 0;
	}

	sched_time_shift_by(hyperperiod);
}

static inline byte idleness_interval_required(unsigned long now)
{
	if (idle_times_idx < idle_times_len) {
		struct idle_time_record next = get_idle_time(idle_times_idx);
		if (time_after(now, next.start)) {
		    	/* yes, we are in an idle interval */
			if (time_after(next.end(), now)) {
				/* idle interval has not yet ended */
				return true;
			} else {
				/* idle interval has ended, move to next */
				idle_times_idx++;
			}
		}
	}
	return false;
}

static void add_to_boosted_jobs_list(byte b)
{
	byte i = 0;
	while (prio_boosted_tasks[i])
		i++;
	prio_boosted_tasks[i] = b;
}

static void remove_from_boosted_jobs_list(byte b)
{
	byte i = 0;
	/* find b */
	while (prio_boosted_tasks[i] != b)
		i++;
	/* now shift the rest */
	while (prio_boosted_tasks[i]) {
		prio_boosted_tasks[i] = prio_boosted_tasks[i + 1];
		i++;
	}
}

static void oe_release_job(byte i)
{
	/* unsigned long: possible overflow accepted and ok */
	next_arrival_time[i] += next_inter_arrival_time[i];
	next_inter_arrival_time[i] = tasks[i].period;

	if (irregs_idx[i] < irregs_len[i]) {
		struct irregularity_record next =
			get_irregularity(i, irregs_idx[i]);
		if (next.job_id == next_job_id[i]) {
			/* special case: next arrival is irregular */
			next_inter_arrival_time[i] -= next.delay;
			next_arrival_time[i] += next.delay;
			/* move to next irregular arrival record */
			irregs_idx[i]++;
			add_to_boosted_jobs_list(i);
		}
	}

	next_job_id[i]++;
}

sched_decision_t oe_schedule(void)
{
	unsigned long now = current_sched_time();

	/* ASSUMPTION: task index zero is the highest-priority task */

	/* first check for hyperperiod crossing */
	if (time_after(now, hyperperiod)) {
		oe_hyperperiod_crossing();
		now -= hyperperiod;
		/* At the beginning of a new hyperperiod, we know that it's
		 * task 0 that has a ready job. Take a shortcut here. */
		oe_release_job(0);
		return (sched_decision_t)
			{tasks[0].job_function, tasks[0].wcet};

	/* otherwise check for idle-time insertion */
	} else if (idleness_interval_required(now))
		/* we need to idle => dispatch no job */
		return NO_JOB_READY;

	/* if we reach here, make a regular scheduling decision */

	/* first check for priority-boosted jobs */
	for (byte b = 0; b < number_of_tasks && prio_boosted_tasks[b]; b++) {
		byte i = prio_boosted_tasks[b];
		if (time_after(now, next_arrival_time[i])) {
			/* a job of this task has been released */
			remove_from_boosted_jobs_list(i);
			oe_release_job(i);
			return (sched_decision_t)
				{tasks[i].job_function, tasks[i].wcet};
		}
	}

	/* finally, look for any regular-priority jobs */
	for (byte i = 0; i < number_of_tasks; i++) {
		if (time_after(now, next_arrival_time[i])) {
			/* a job of this task has been released */
			oe_release_job(i);
			return (sched_decision_t)
				{tasks[i].job_function, tasks[i].wcet};
		}
	}

	/* If we reach here, no job is ready. Just idle. */
	return NO_JOB_READY;
}
