/* An implementation of non-preemptive, non-work-conserving
 * periodic pre-cautious RM scheduling. */

#include "Arduino.h"

#include "tasks.h"
#include "sched_time.h"

static unsigned long next_release_time[MAX_TASKS];

void pcrm_init(void)
{
	for (int i = 0; i < MAX_TASKS; i++)
		next_release_time[i] = 0UL;
}

static void pcrm_release_job(byte i)
{
	/* unsigned long: possible overflow accepted and ok */
	next_release_time[i] += tasks[i].period;
}

sched_decision_t pcrm_schedule(void)
{
	unsigned long now = current_sched_time();

	/* ASSUMPTION: task index zero is the highest-priority task and
	 *             has the shortest period. */

	/* check task_0 first */
	if (time_after(now, next_release_time[0])) {
		/* a job of task_0 has been released */
		pcrm_release_job(0);
		return (sched_decision_t)
			{tasks[0].job_function, tasks[0].wcet};
	}

	/* if task_0 is not pending, check the rest and test the pre-cautious
	 * RM condition.*/
	for (byte i = 1; i < number_of_tasks; i++)
	{
		if (time_after(now, next_release_time[i])) {
			/* a job of this task could be released */

			/* first check that it will not induce a deadline miss
			 * in task_0 (the pre-cautious RM condition) */
			if (time_after(now + tasks[i].wcet + tasks[0].wcet,
			               next_release_time[0] + tasks[0].deadline)) {
			        /* it fits, release and execute it */
				pcrm_release_job(i);
				return (sched_decision_t)
					{tasks[i].job_function, tasks[i].wcet};
			}
		}
	}

	/* If we reach here, no job is ready, or we need to wait for a
	 * release of a job of task_0. Just idle. */
	return NO_JOB_READY;
}
